# Generated by Django 3.2.3 on 2022-07-30 09:50

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("tests", "0005_customsoftdelete"),
    ]

    operations = [
        migrations.CreateModel(
            name="CustomTableName",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("int_field", models.IntegerField(null=True, unique=True)),
            ],
            options={
                "db_table": "order",
            },
        ),
    ]
