# Generated by Django 5.2 on 2025-04-18 08:55

import django
from django.db import migrations, models


def get_operations():
    if django.VERSION >= (5, 2):
        return [
            migrations.CreateModel(
                name="SoftDeleteCompositePk",
                fields=[
                    ("id_1", models.IntegerField()),
                    ("id_2", models.IntegerField()),
                    (
                        "pk",
                        models.CompositePrimaryKey(
                            "id_1",
                            "id_2",
                            blank=True,
                            editable=False,
                            primary_key=True,
                            serialize=False,
                        ),
                    ),
                    ("is_active", models.BooleanField(default=True)),
                    ("other_field", models.TextField()),
                ],
            ),
        ]
    else:  # pragma: no cover
        return []


class Migration(migrations.Migration):
    dependencies = [
        ("tests", "0013_alter_testtrigger_m2m_field_changedcondition"),
    ]

    operations = get_operations()
