.md-typeset__table {
  min-width: 100%;
}

.md-typeset table:not([class]) {
  display: table;
}

:root {
  --md-primary-fg-color: #1d1f29;
  --md-primary-fg-color--light: #1d1f29;
  --md-primary-fg-color--dark: #1d1f29;
}

.md-content {
--md-typeset-a-color: #00bc70;
}

.md-footer {
  background-color: #1d1f29;
}
.md-footer-meta {
  background-color: #1d1f29;
}

readthedocs-flyout {
  display: none;
}