# Generated by Django 3.0.7 on 2020-07-21 19:46

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("tests", "0003_auto_20200718_0938"),
    ]

    operations = [
        migrations.CreateModel(
            name="FSM",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("transition", models.CharField(max_length=32)),
            ],
        ),
    ]
