# Generated by Django 4.2.6 on 2023-10-11 20:50

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("tests", "0012_alter_partitionmodel_options"),
    ]

    operations = [
        migrations.AlterField(
            model_name="testtrigger",
            name="m2m_field",
            field=models.ManyToManyField(related_name="+", to=settings.AUTH_USER_MODEL),
        ),
        migrations.CreateModel(
            name="ChangedCondition",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("field", models.CharField(max_length=16)),
                ("int_field", models.IntegerField(default=0)),
                ("dt_field", models.DateTimeField(auto_now=True)),
                ("nullable", models.CharField(default=None, max_length=16, null=True)),
                (
                    "char_pk_fk_field",
                    models.ForeignKey(
                        null=True, on_delete=django.db.models.deletion.CASCADE, to="tests.charpk"
                    ),
                ),
                (
                    "fk_field",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "m2m_field",
                    models.ManyToManyField(related_name="+", to=settings.AUTH_USER_MODEL),
                ),
            ],
        ),
    ]
