# Generated by Django 3.2.15 on 2022-08-08 10:55

from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("tests", "0006_customtablename"),
    ]

    operations = [
        migrations.CreateModel(
            name="TestTriggerProxy",
            fields=[],
            options={
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=("tests.testtrigger",),
        ),
        migrations.AddField(
            model_name="testtrigger",
            name="m2m_field",
            field=models.ManyToManyField(
                related_name="_tests_testtrigger_m2m_field_+", to=settings.AUTH_USER_MODEL
            ),
        ),
        migrations.CreateModel(
            name="TestDefaultThrough",
            fields=[],
            options={
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=("tests.testtrigger_m2m_field",),
        ),
    ]
